@@ -210,7 +210,8 @@ typedef enum {
 } GtkHTMLCursorSkipType;
 
 typedef enum {
-	GTK_HTML_EDITOR_EVENT_COMMAND,
+	GTK_HTML_EDITOR_EVENT_COMMAND_BEFORE,
+	GTK_HTML_EDITOR_EVENT_COMMAND_AFTER,
 	GTK_HTML_EDITOR_EVENT_IMAGE_URL,
 	GTK_HTML_EDITOR_EVENT_DELETE
 } GtkHTMLEditorEventType;
