@@ -33,19 +33,23 @@
 GtkHTMLClassProperties *
 gtk_html_class_properties_new (void)
 {
-	GtkHTMLClassProperties *p = g_new (GtkHTMLClassProperties, 1);
+	GtkHTMLClassProperties *p = g_new0 (GtkHTMLClassProperties, 1);
 
 	/* default values */
 	p->magic_links             = TRUE;
-	p->keybindings_theme       = g_strdup ("emacs");
-	p->font_var         = g_strdup ("-*-helvetica-*-*-*-*-12-*-*-*-*-*-*-*");
-	p->font_fix         = g_strdup ("-*-courier-*-*-*-*-12-*-*-*-*-*-*-*");
+	p->keybindings_theme       = g_strdup ("ms");
+	p->font_var                = g_strdup ("-*-helvetica-*-*-*-*-12-*-*-*-*-*-*-*");
+	p->font_fix                = g_strdup ("-*-courier-*-*-*-*-12-*-*-*-*-*-*-*");
 	p->font_var_size           = DEFAULT_FONT_SIZE;
 	p->font_fix_size           = DEFAULT_FONT_SIZE;
-	p->font_var_print   = g_strdup ("-*-helvetica-*-*-*-*-12-*-*-*-*-*-*-*");
-	p->font_fix_print   = g_strdup ("-*-courier-*-*-*-*-12-*-*-*-*-*-*-*");
+	p->font_var_points         = FALSE;
+	p->font_fix_points         = FALSE;
+	p->font_var_print          = g_strdup ("-*-helvetica-*-*-*-*-12-*-*-*-*-*-*-*");
+	p->font_fix_print          = g_strdup ("-*-courier-*-*-*-*-12-*-*-*-*-*-*-*");
 	p->font_var_size_print     = DEFAULT_FONT_SIZE;
 	p->font_fix_size_print     = DEFAULT_FONT_SIZE;
+	p->font_var_print_points   = FALSE;
+	p->font_fix_print_points   = FALSE;
 	p->animations              = TRUE;
 
 	p->live_spell_check        = TRUE;
@@ -90,12 +94,16 @@ gtk_html_class_properties_load (GtkHTMLClassProperties *p, GConfClient *client)
 	     g_free (p->font_fix), g_strdup);
 	GET (int, "/font_variable_size", font_var_size,,);
 	GET (int, "/font_fixed_size", font_fix_size,,);
+	GET (bool, "/font_variable_points", font_var_points,,);
+	GET (bool, "/font_fixed_points", font_fix_points,,);
 	GET (string, "/font_variable_print", font_var_print,
 	     g_free (p->font_var_print), g_strdup);
 	GET (string, "/font_fixed_print", font_fix_print,
 	     g_free (p->font_fix_print), g_strdup);
 	GET (int, "/font_variable_size_print", font_var_size_print,,);
 	GET (int, "/font_fixed_size_print", font_fix_size_print,,);
+	GET (bool, "/font_variable_print_points", font_var_print_points,,);
+	GET (bool, "/font_fixed_print_points", font_fix_print_points,,);
 
 	GET (bool, "/live_spell_check", live_spell_check,,);
 	GET (int, "/spell_error_color_red",   spell_error_color.red,,);
@@ -125,17 +133,25 @@ gtk_html_class_properties_update (GtkHTMLClassProperties *p, GConfClient *client
 		SET (string, "/font_variable", font_var);
 	if (strcmp (p->font_fix, old->font_fix))
 		SET (string, "/font_fixed", font_fix);
-	if (p->font_var_size != old->font_var_size)
+	if (p->font_var_points != old->font_var_points)
+		SET (bool, "/font_variable_points", font_var_points);
+	if (p->font_fix_points != old->font_fix_points)
+		SET (bool, "/font_fixed_points", font_fix_points);
+	if (p->font_var_size != old->font_var_size || p->font_var_points != old->font_var_points)
 		SET (int, "/font_variable_size", font_var_size);
-	if (p->font_fix_size != old->font_fix_size)
+	if (p->font_fix_size != old->font_fix_size || p->font_fix_points != old->font_fix_points)
 		SET (int, "/font_fixed_size", font_fix_size);
 	if (strcmp (p->font_var_print, old->font_var_print))
 		SET (string, "/font_variable_print", font_var_print);
 	if (strcmp (p->font_fix_print, old->font_fix_print))
 		SET (string, "/font_fixed_print", font_fix_print);
-	if (p->font_var_size_print != old->font_var_size_print)
+	if (p->font_var_print_points != old->font_var_print_points)
+		SET (bool, "/font_variable_print_points", font_var_print_points);
+	if (p->font_fix_print_points != old->font_fix_print_points)
+		SET (bool, "/font_fixed_print_points", font_fix_print_points);
+	if (p->font_var_size_print != old->font_var_size_print || p->font_var_print_points != old->font_var_print_points)
 		SET (int, "/font_variable_size_print", font_var_size_print);
-	if (p->font_fix_size_print != old->font_fix_size_print)
+	if (p->font_fix_size_print != old->font_fix_size_print || p->font_fix_print_points != old->font_fix_print_points)
 		SET (int, "/font_fixed_size_print", font_fix_size_print);
 
 	if (p->live_spell_check != old->live_spell_check)
@@ -189,8 +205,13 @@ gtk_html_class_properties_load (GtkHTMLClassProperties *p)
 	GET  (int, font_fix_size_print, s);
 	g_free (s);
 
+	GET  (bool, font_var_points, "font_variable_points=false");
+	GET  (bool, font_fix_points, "font_fixed_points=false");
+	GET  (bool, font_var_print_points, "font_variable_print_points=false");
+	GET  (bool, font_fix_print_points, "font_fixed_print_points=false");
+
 	GET  (bool, live_spell_check, "live_spell_check=true");
-	GET  (int, spell_error_color.red,   "spell_error_color_red=0xffff");
+	GET  (int, spell_error_color.red,   "spell_error_color_red=65535");
 	GET  (int, spell_error_color.green, "spell_error_color_green=0");
 	GET  (int, spell_error_color.blue,  "spell_error_color_blue=0");
 	GETS (language, "language=en");
@@ -213,6 +234,10 @@ gtk_html_class_properties_save (GtkHTMLClassProperties *p)
 	gnome_config_set_string ("font_fixed_print", p->font_fix_print);
 	gnome_config_set_int ("font_variable_size_print", p->font_var_size_print);
 	gnome_config_set_int ("font_fixed_size_print", p->font_fix_size_print);
+	gnome_config_set_bool ("font_variable_points", p->font_var_points);
+	gnome_config_set_bool ("font_fixed_points", p->font_fix_points);
+	gnome_config_set_bool ("font_variable_print_points", p->font_var_print_points);
+	gnome_config_set_bool ("font_fixed_print_points", p->font_fix_print_points);
 
 	gnome_config_set_bool ("live_spell_check", p->live_spell_check);
 	gnome_config_set_int  ("spell_error_color_red",   p->spell_error_color.red);
@@ -245,6 +270,10 @@ gtk_html_class_properties_copy (GtkHTMLClassProperties *p1,
 	COPYS (font_fix_print);
 	COPY  (font_var_size_print);
 	COPY  (font_fix_size_print);
+	COPY  (font_var_points);
+	COPY  (font_fix_points);
+	COPY  (font_var_print_points);
+	COPY  (font_fix_print_points);
 
 	COPY  (live_spell_check);
 	COPY  (spell_error_color);
@@ -276,6 +305,7 @@ static GtkEnumValue _gtk_html_command_values[] = {
   { GTK_HTML_COMMAND_UNDO,  "GTK_HTML_COMMAND_UNDO",  "undo" },
   { GTK_HTML_COMMAND_REDO,  "GTK_HTML_COMMAND_REDO",  "redo" },
   { GTK_HTML_COMMAND_COPY,  "GTK_HTML_COMMAND_COPY",  "copy" },
+  { GTK_HTML_COMMAND_COPY_AND_DISABLE_SELECTION,  "GTK_HTML_COMMAND_COPY_AND_DISABLE_SELECTION",  "copy-and-disable-selection" },
   { GTK_HTML_COMMAND_CUT,   "GTK_HTML_COMMAND_CUT",   "cut" },
   { GTK_HTML_COMMAND_PASTE, "GTK_HTML_COMMAND_PASTE", "paste" },
   { GTK_HTML_COMMAND_CUT_LINE, "GTK_HTML_COMMAND_CUT_LINE", "cut-line" },
@@ -283,8 +313,8 @@ static GtkEnumValue _gtk_html_command_values[] = {
   { GTK_HTML_COMMAND_INSERT_RULE, "GTK_HTML_COMMAND_INSERT_RULE", "insert-rule" },
   { GTK_HTML_COMMAND_INSERT_PARAGRAPH, "GTK_HTML_COMMAND_INSERT_PARAGRAPH", "insert-paragraph" },
   { GTK_HTML_COMMAND_INSERT_TAB, "GTK_HTML_COMMAND_INSERT_TAB", "insert-tab" },
-  { GTK_HTML_COMMAND_INSERT_TAB_OR_INDENT_MORE_OR_NEXT_CELL,
-    "GTK_HTML_COMMAND_INSERT_TAB_OR_INDENT_MORE_OR_NEXT_CELL", "insert-tab-or-indent-more-or-next-cell" },
+  { GTK_HTML_COMMAND_INSERT_TAB_OR_NEXT_CELL,
+    "GTK_HTML_COMMAND_INSERT_TAB_OR_NEXT_CELL", "insert-tab-or-next-cell" },
   { GTK_HTML_COMMAND_DELETE, "GTK_HTML_COMMAND_DELETE", "delete" },
   { GTK_HTML_COMMAND_DELETE_BACK, "GTK_HTML_COMMAND_DELETE_BACK", "delete-back" },
   { GTK_HTML_COMMAND_DELETE_BACK_OR_INDENT_DEC, "GTK_HTML_COMMAND_DELETE_BACK_OR_INDENT_DEC", "delete-back-or-indent-dec" },
@@ -316,8 +346,9 @@ static GtkEnumValue _gtk_html_command_values[] = {
   { GTK_HTML_COMMAND_ALIGN_RIGHT, "GTK_HTML_COMMAND_ALIGN_RIGHT", "align-right" },
   { GTK_HTML_COMMAND_INDENT_ZERO, "GTK_HTML_COMMAND_INDENT_ZERO", "indent-zero" },
   { GTK_HTML_COMMAND_INDENT_INC, "GTK_HTML_COMMAND_INDENT_INC", "indent-more" },
+  { GTK_HTML_COMMAND_INDENT_INC_OR_NEXT_CELL, "GTK_HTML_COMMAND_INDENT_INC_OR_NEXT_CELL", "indent-more-or-next-cell" },
   { GTK_HTML_COMMAND_INDENT_DEC, "GTK_HTML_COMMAND_INDENT_DEC", "indent-less" },
-  { GTK_HTML_COMMAND_INDENT_DEC_OR_PREV_CELL, "GTK_HTML_COMMAND_INDENT_DEC_OR_PREV_CELL", "indent-less-or-prev-cell" },
+  { GTK_HTML_COMMAND_PREV_CELL, "GTK_HTML_COMMAND_PREV_CELL", "prev-cell" },
   { GTK_HTML_COMMAND_INDENT_PARAGRAPH, "GTK_HTML_COMMAND_INDENT_PARAGRAPH", "indent-paragraph" },
   { GTK_HTML_COMMAND_BREAK_AND_FILL_LINE, "GTK_HTML_COMMAND_BREAK_AND_FILL_LINE", "break-and-fill" },
   { GTK_HTML_COMMAND_SPACE_AND_FILL_LINE, "GTK_HTML_COMMAND_SPACE_AND_FILL_LINE", "space-and-fill" },
@@ -333,6 +364,7 @@ static GtkEnumValue _gtk_html_command_values[] = {
   { GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMDOTTED, "GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMDOTTED", "style-itemdot" },
   { GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMROMAN, "GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMROMAN", "style-itemroman" },
   { GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMDIGIT, "GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMDIGIT", "style-itemdigit" },
+  { GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMALPHA, "GTK_HTML_COMMAND_PARAGRAPH_STYLE_ITEMALPHA", "style-itemalpha" },
   { GTK_HTML_COMMAND_MODIFY_SELECTION_UP, "GTK_HTML_COMMAND_MODIFY_SELECTION_UP", "selection-move-up" },
   { GTK_HTML_COMMAND_MODIFY_SELECTION_DOWN, "GTK_HTML_COMMAND_MODIFY_SELECTION_DOWN", "selection-move-down" },
   { GTK_HTML_COMMAND_MODIFY_SELECTION_LEFT, "GTK_HTML_COMMAND_MODIFY_SELECTION_LEFT", "selection-move-left" },
@@ -404,6 +436,7 @@ static GtkEnumValue _gtk_html_command_values[] = {
   { GTK_HTML_COMMAND_DELETE_TABLE_ROW, "GTK_HTML_COMMAND_DELETE_TABLE_ROW", "delete-table-row" },
   { GTK_HTML_COMMAND_DELETE_TABLE_COLUMN, "GTK_HTML_COMMAND_DELETE_TABLE_COLUMN", "delete-table-column" },
   { GTK_HTML_COMMAND_DELETE_TABLE_CELL_CONTENTS, "GTK_HTML_COMMAND_DELETE_TABLE_CELL_CONTENTS", "delete-cell-contents" },
+  { GTK_HTML_COMMAND_GRAB_FOCUS, "GTK_HTML_COMMAND_GRAB_FOCUS", "grab-focus" },
   { 0, NULL, NULL }
 };
 
