@@ -34,6 +34,7 @@
 #include "htmlcolorset.h"
 #include "htmlengine.h"
 #include "htmlengine-edit.h"
+#include "htmlengine-edit-cursor.h"
 #include "htmlengine-edit-cut-and-paste.h"
 #include "htmlengine-edit-fontstyle.h"
 #include "htmlengine-edit-movement.h"
@@ -48,13 +49,17 @@
 #include "htmlundo.h"
 #include "htmlundo-action.h"
 
-static void        delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoDirection dir);
-static void        insert_object (HTMLEngine *e, HTMLObject *obj, guint len, HTMLUndoDirection dir, gboolean check);
+static gint        delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoDirection dir,
+				  gboolean add_prop);
+static void        insert_object (HTMLEngine *e, HTMLObject *obj, guint len, guint position_after, gint level,
+				  HTMLUndoDirection dir, gboolean check);
 static void        append_object (HTMLEngine *e, HTMLObject *o, guint len, HTMLUndoDirection dir);
-static void        insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir);
+static void        insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir, gboolean add_undo);
 
 /* helper functions -- need refactor */
 
+/* #define OP_DEBUG */
+
 static void
 html_cursor_get_left (HTMLCursor *cursor, HTMLObject **obj, gint *off)
 {
@@ -69,20 +74,6 @@ html_cursor_get_left (HTMLCursor *cursor, HTMLObject **obj, gint *off)
 	*off = cursor->offset;
 }
 
-static void
-html_cursor_get_right (HTMLCursor *cursor, HTMLObject **obj, gint *off)
-{
-	if (cursor->offset >= html_object_get_length (cursor->object)) {
-		*obj = html_object_next_not_slave (cursor->object);
-		if (*obj) {
-			*off = 0;
-		}
-		return;
-	}
-	*obj = cursor->object;
-	*off = cursor->offset;
-}
-
 static void
 html_point_get_left (HTMLPoint *source, HTMLPoint *dest)
 {
@@ -182,12 +173,13 @@ get_common_parent (HTMLObject *from, HTMLObject *to)
 	return parent ? parent : try_find_common_parent_of (to, from);
 }
 
-static void
+static gint
 prepare_delete_bounds (HTMLEngine *e, GList **from_list, GList **to_list,
 		       GList **bound_left, GList **bound_right)
 {
 	HTMLPoint b_left, b_right, begin, end;
 	HTMLObject *common_parent;
+	gint ret_level;
 
 	g_assert (e->selection);
 
@@ -195,6 +187,7 @@ prepare_delete_bounds (HTMLEngine *e, GList **from_list, GList **to_list,
 	html_point_get_left  (&e->selection->to,   &end);
 
 	common_parent = get_common_parent (begin.object, end.object);
+	ret_level     = get_parent_depth (begin.object, common_parent);
 
 	*from_list = point_get_parent_list (&begin, get_parent_depth (begin.object, common_parent), TRUE);
 	*to_list   = point_get_parent_list (&end,   get_parent_depth (end.object, common_parent),   TRUE);
@@ -217,15 +210,17 @@ prepare_delete_bounds (HTMLEngine *e, GList **from_list, GList **to_list,
 		if (level > 1 && *bound_right)
 			*bound_right = g_list_prepend (*bound_right, NULL);
 	}
+
+	return ret_level;
 }
 
 static void
 remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right, HTMLCursor *c)
 {
 	HTMLObject *lo, *ro, *prev;
-	gint len;
 
-	/* printf ("before merge\n");
+#ifdef OP_DEBUG
+	printf ("before merge\n");
 	gtk_html_debug_dump_tree_simple (e->clue, 0);
 	if (left && left->data) {
 		printf ("left\n");
@@ -235,13 +230,12 @@ remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right
 	if (right && right->data) {
 		printf ("right\n");
 		gtk_html_debug_dump_tree_simple (right->data, 0);
-		} */
-
+	}
+#endif
 	while (left && left->data && right && right->data) {
 
 		lo  = HTML_OBJECT (left->data);
 		ro  = HTML_OBJECT (right->data);
-		len = html_object_get_length (lo);
 
 		left  = left->next;
 		right = right->next;
@@ -249,10 +243,14 @@ remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right
 		if (html_object_is_text (lo) && !*HTML_TEXT (lo)->text && (html_object_prev_not_slave (lo) || merge)) {
 			HTMLObject *nlo = html_object_prev_not_slave (lo);
 
-			if (e->cursor->object == lo)
+			if (e->cursor->object == lo) {
 				e->cursor->object = ro;
-			if (c && c->object == lo)
+				e->cursor->offset = 0;
+			}
+			if (c && c->object == lo) {
 				c->object = ro;
+				c->offset = 0;
+			}
 
 			html_object_remove_child (lo->parent, lo);
 			html_object_destroy (lo);
@@ -260,8 +258,10 @@ remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right
 		} else if (html_object_is_text (ro) && !*HTML_TEXT (ro)->text && (html_object_next_not_slave (ro) || merge)) {
 			HTMLObject *nro = html_object_next_not_slave (ro);
 
-			if (e->cursor->object == ro)
+			if (e->cursor->object == ro) {
 				e->cursor->object = lo;
+				e->cursor->offset = html_object_get_length (lo);
+			}
 
 			html_object_remove_child (ro->parent, ro);
 			html_object_destroy (ro);
@@ -269,11 +269,11 @@ remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right
 		}
 
 		if (merge && lo && ro) {
-			if (!html_object_merge (lo, ro, e, left, right))
+			if (!html_object_merge (lo, ro, e, &left, &right, c))
 				break;
 			if (ro == e->cursor->object) {
 				e->cursor->object  = lo;
-				e->cursor->offset += len;
+				e->cursor->offset += html_object_get_length (lo);;
 			}
 		}
 	}
@@ -283,99 +283,28 @@ remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right
 		e->cursor->object = prev;
 		e->cursor->offset = html_object_get_length (e->cursor->object);
 	}
-	/* printf ("-- after\n");
+#ifdef OP_DEBUG
+	printf ("-- after\n");
 	gtk_html_debug_dump_tree_simple (e->clue, 0);
-	printf ("-- END merge\n"); */
-}
-
-static gboolean
-look_for_non_appendable (GList *list)
-{
-	while (list) {
-		HTMLObject *o = HTML_OBJECT (list->data);
-
-		if (o->parent && HTML_OBJECT_TYPE (o->parent) == HTML_TYPE_TABLE)
-			return TRUE;
-		list = list->next;
-	}
-
-	return FALSE;
-}
-
-static HTMLObject *
-split_between_objects (HTMLObject *object, HTMLObject *other,
-		       GList **object_parents, GList **other_parents, gint *object_level)
-{
-	HTMLObject *common_parent;
-	gint d_object, d_other;
-
-	common_parent = get_common_parent (object, other);
-
-	if (common_parent) {
-		d_object = get_parent_depth (object, common_parent);
-		d_other  = get_parent_depth (other,  common_parent);
-
-		if (d_object <= *object_level) {
-			object_get_parent_list (object, d_object - 1, object_parents);
-			object_get_parent_list (other, d_other - 1, other_parents);
-
-#ifdef GTKHTML_DEBUG_TABLE
-			if (*object_parents) {
-				printf ("split object\n");
-				gtk_html_debug_dump_tree_simple ((*object_parents)->data, 0);
-			}
-
-			if (*other_parents) {
-				printf ("split other\n");
-				gtk_html_debug_dump_tree_simple ((*other_parents)->data, 0);
-			}
+	printf ("-- END merge\n");
 #endif
-			if (!look_for_non_appendable (*object_parents) && !look_for_non_appendable (*object_parents)) {
-				g_list_free (*object_parents);
-				g_list_free (*other_parents);
-
-				printf ("drop lists, do simple split\n");
-				*object_parents = NULL;
-				*other_parents = NULL;
-
-				return NULL;
-			}
-
-			*object_level -= d_object - 1;
-
-			return common_parent;
-		}
-	}
-
-	return NULL;
 }
 
 static void
 split_and_add_empty_texts (HTMLEngine *e, gint level, GList **left, GList **right)
 {
-	HTMLObject *object = NULL;
-
-	if ((e->cursor->offset == 0 || (html_object_get_length (e->cursor->object) == e->cursor->offset))) {
-		HTMLObject *leaf;
-		gboolean prev;
-
-		prev = html_object_get_length (e->cursor->object) != e->cursor->offset;
-
-		leaf = prev
-		? html_object_prev_leaf_not_type (e->cursor->object, HTML_TYPE_TEXTSLAVE)
-		: html_object_next_leaf_not_type (e->cursor->object, HTML_TYPE_TEXTSLAVE);
-		if (leaf) {
-			if (prev)
-				object = split_between_objects (e->cursor->object, leaf, right, left, &level);
-			else
-				object = split_between_objects (e->cursor->object, leaf, left, right, &level);
-		}
-	}
-
-	if (!object)
-		object = e->cursor->object;
-
-	html_object_split (object, e, *right ? HTML_OBJECT ((*right)->data) : NULL, e->cursor->offset, level, left, right);
+#ifdef OP_DEBUG
+	printf ("-- SPLIT begin\n");
+	gtk_html_debug_dump_tree_simple (e->clue, 0);
+	printf ("-- SPLIT middle\n");
+#endif
+	html_object_split (e->cursor->object, e, *right ? HTML_OBJECT ((*right)->data) : NULL,
+			   e->cursor->offset, level, left, right);
+#ifdef OP_DEBUG
+	printf ("-- SPLIT middle\n");
+	gtk_html_debug_dump_tree_simple (e->clue, 0);
+	printf ("-- SPLIT end\n");
+#endif
 }
 
 /* end of helper */
@@ -386,16 +315,14 @@ html_engine_copy_object (HTMLEngine *e, HTMLObject **o, guint *len)
 	GList *from, *to;
 
 	if (html_engine_is_selection_active (e)) {
-		html_engine_freeze (e);
 		prepare_delete_bounds (e, &from, &to, NULL, NULL);
 		*len = 0;
 		*o    = html_object_op_copy (HTML_OBJECT (from->data), e,
-				from->next, to->next, len);
-#ifdef GTKHTML_DEBUG_TABLE
+					     from->next, to->next, len);
+#ifdef OP_DEBUG
 		printf ("copy len: %d (parent %p)\n", *len, (*o)->parent);
 		gtk_html_debug_dump_tree_simple (*o, 0);
 #endif
-		html_engine_thaw (e);
 	}
 }
 
@@ -410,6 +337,7 @@ struct _DeleteUndo {
 
 	HTMLObject *buffer;
 	guint       buffer_len;
+	gint        level;
 };
 typedef struct _DeleteUndo DeleteUndo;
 
@@ -423,7 +351,7 @@ delete_undo_destroy (HTMLUndoData *data)
 }
 
 static void
-delete_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir)
+delete_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir, guint position_after)
 {
 	DeleteUndo *undo;
 	HTMLObject *buffer;
@@ -431,11 +359,11 @@ delete_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir)
 
 	undo         = (DeleteUndo *) data;
 	buffer       = html_object_op_copy (undo->buffer, e, NULL, NULL, &len);
-	insert_object (e, buffer, undo->buffer_len, html_undo_direction_reverse (dir), TRUE);
+	insert_object (e, buffer, undo->buffer_len, position_after, undo->level, html_undo_direction_reverse (dir), TRUE);
 }
 
 static void
-delete_setup_undo (HTMLEngine *e, HTMLObject *buffer, guint len, HTMLUndoDirection dir)
+delete_setup_undo (HTMLEngine *e, HTMLObject *buffer, guint len, guint position_after, gint level, HTMLUndoDirection dir)
 {
 	DeleteUndo *undo;
 
@@ -445,12 +373,14 @@ delete_setup_undo (HTMLEngine *e, HTMLObject *buffer, guint len, HTMLUndoDirecti
 	undo->data.destroy = delete_undo_destroy;
 	undo->buffer       = buffer;
 	undo->buffer_len   = len;
+	undo->level        = level;
 
 	/* printf ("delete undo len %d\n", len); */
 
 	html_undo_add_action (e->undo,
 			      html_undo_action_new ("Delete", delete_undo_action,
-						    HTML_UNDO_DATA (undo), html_cursor_get_position (e->cursor)),
+						    HTML_UNDO_DATA (undo), html_cursor_get_position (e->cursor),
+						    position_after),
 			      dir);
 }
 
@@ -467,15 +397,6 @@ move_cursor_before_delete (HTMLEngine *e)
 				e->cursor->object = obj;
 				e->cursor->offset = off;
 			}
-		} else {
-			HTMLObject *obj;
-			gint off;
-
-			html_cursor_get_right (e->mark, &obj, &off);
-			if (obj) {
-				e->cursor->object = obj;
-				e->cursor->offset = 0;
-			}
 		}
 	}
 }
@@ -492,21 +413,27 @@ place_cursor_before_mark (HTMLEngine *e)
 	}
 }
 
-static void
+static gint
 delete_object_do (HTMLEngine *e, HTMLObject **object, guint *len)
 {
 	GList *from, *to, *left, *right;
+	guint position;
+	gint level;
 
 	html_engine_freeze (e);
-	prepare_delete_bounds (e, &from, &to, &left, &right);
+	level = prepare_delete_bounds (e, &from, &to, &left, &right);
 	place_cursor_before_mark (e);
 	move_cursor_before_delete (e);
 	html_engine_disable_selection (e);
 	*len     = 0;
 	*object  = html_object_op_cut  (HTML_OBJECT (from->data), e, from->next, to->next, left, right, len);
+	position = e->cursor->position;
 	remove_empty_and_merge (e, TRUE, left ? left->next : NULL, right ? right->next : NULL, NULL);
+	e->cursor->position = position;
 	html_engine_spell_check_range (e, e->cursor, e->cursor);
 	html_engine_thaw (e);
+
+	return level;
 }
 
 static void
@@ -531,13 +458,14 @@ check_table_1 (HTMLEngine *e)
 		html_cursor_forward (head, e);
 }
 
-static void
-delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoDirection dir)
+static gint
+delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoDirection dir, gboolean add_undo)
 {
 	html_engine_edit_selection_updater_update_now (e->selection_updater);
 	if (html_engine_is_selection_active (e)) {
 		HTMLObject *object;
-		guint len;
+		guint len, position_before;
+		gint level;
 
 		if (!html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))
 		    && !html_clueflow_is_empty (HTML_CLUEFLOW (e->mark->object->parent))) {
@@ -546,34 +474,46 @@ delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoD
 		}
 		if (e->cursor->position == e->mark->position) {
 			html_engine_disable_selection (e);
-			return;
+			return 0;
 		}
-		delete_object_do (e, &object, &len);
+		position_before = MAX (e->cursor->position, e->mark->position);
+		level = delete_object_do (e, &object, &len);
 		if (ret_object && ret_len) {
 			*ret_object = html_object_op_copy (object, e, NULL, NULL, ret_len);
 			*ret_len    = len;
 		}
-		delete_setup_undo (e, object, len, dir);
+		if (add_undo) {
+			delete_setup_undo (e, object, len, position_before, level, dir);
+		} else
+			html_object_destroy (object);
 		gtk_html_editor_event (e->widget, GTK_HTML_EDITOR_EVENT_DELETE, NULL);
+
+		return level;
 	}
+
+	return 0;
 }
 
 void
 html_engine_delete (HTMLEngine *e)
 {
-	delete_object (e, NULL, NULL, HTML_UNDO_UNDO);
+	delete_object (e, NULL, NULL, HTML_UNDO_UNDO, TRUE);
 }
 
-void
+gint
 html_engine_cut (HTMLEngine *e)
 {
-	html_engine_clipboard_clear (e);
-	delete_object (e, &e->clipboard, &e->clipboard_len, HTML_UNDO_UNDO);
+	gint rv;
 
-#ifdef GTKHTML_DEBUG_TABLE
+	html_engine_clipboard_clear (e);
+	rv = delete_object (e, &e->clipboard, &e->clipboard_len, HTML_UNDO_UNDO, TRUE);
+
+#ifdef OP_DEBUG
 	printf ("cut  len: %d\n", e->clipboard_len);
 	gtk_html_debug_dump_tree_simple (e->clipboard, 0);
 #endif
+
+	return rv;
 }
 
 /*
@@ -584,57 +524,63 @@ static void
 set_cursor_at_end_of_object (HTMLEngine *e, HTMLObject *o, guint len)
 {
 	guint save_position;
+	gboolean need_spell_check;
 
 	save_position       = e->cursor->position;
 	e->cursor->object   = html_object_get_tail_leaf (o);
+	need_spell_check = e->need_spell_check;
+	e->need_spell_check = FALSE;
 	while (html_cursor_forward (e->cursor, e))
 		;
+	e->need_spell_check = need_spell_check;
 	e->cursor->position = save_position + len;
 	e->cursor->offset   = html_object_get_length (e->cursor->object);
 }
 
 static inline void
-insert_object_do (HTMLEngine *e, HTMLObject *obj, guint len, gboolean check, HTMLUndoDirection dir)
+isolate_tables (HTMLEngine *e, HTMLUndoDirection dir, guint position_before, guint position_after,
+		  gboolean *delete_paragraph_before, gboolean *delete_paragraph_after)
+{
+	HTMLObject *next;
+
+	*delete_paragraph_after  = FALSE;
+	*delete_paragraph_before = FALSE;
+
+	html_cursor_jump_to_position_no_spell (e->cursor, e, position_after);
+	next = html_object_next_not_slave (e->cursor->object);
+	if (next && e->cursor->offset == html_object_get_length (e->cursor->object)
+	    && (HTML_IS_TABLE (e->cursor->object) || HTML_IS_TABLE (next))) {
+		insert_empty_paragraph (e, dir, FALSE);
+		*delete_paragraph_after = TRUE;
+	}
+
+	html_cursor_jump_to_position_no_spell (e->cursor, e, position_before);
+	next = html_object_next_not_slave (e->cursor->object);
+	if (next && e->cursor->offset == html_object_get_length (e->cursor->object)
+	    && (HTML_IS_TABLE (e->cursor->object) || HTML_IS_TABLE (next))) {
+		insert_empty_paragraph (e, dir, FALSE);
+		*delete_paragraph_before = TRUE;
+	}
+}
+
+static inline void
+insert_object_do (HTMLEngine *e, HTMLObject *obj, guint *len, gint level, guint position_after,
+		  gboolean check, HTMLUndoDirection dir)
 {
-	HTMLObject *cur;
 	HTMLCursor *orig;
 	GList *left = NULL, *right = NULL;
 	GList *first = NULL, *last = NULL;
-	gint level;
+	guint position_before;
 
 	html_engine_freeze (e);
-
-	if (HTML_IS_TABLE (e->cursor->object)) {
-		if (e->cursor->offset) {
-			HTMLObject *head = html_object_get_head_leaf (obj);
-
-			if (!head->parent || (HTML_IS_CLUEFLOW (head->parent)
-					      && !html_clueflow_is_empty (HTML_CLUEFLOW (head->parent))))
-				insert_empty_paragraph (e, dir);
-		} else {
-			HTMLObject *tail = html_object_get_tail_leaf (obj);
-
-			if (!tail->parent || (HTML_IS_CLUEFLOW (tail->parent)
-					      && !html_clueflow_is_empty (HTML_CLUEFLOW (tail->parent)))) {
-				insert_empty_paragraph (e, dir);
-				html_cursor_backward (e->cursor, e);
-			}
-		}
-	}
-
-	level = 0;
-	cur   = html_object_get_head_leaf (obj);
-	while (cur) {
-		level++;
-		cur = cur->parent;
-	}
-	orig = html_cursor_dup (e->cursor);
-
+	position_before = e->cursor->position;
 	html_object_change_set_down (obj, HTML_CHANGE_ALL);
-	split_and_add_empty_texts (e, MIN (3, level), &left, &right);
+	split_and_add_empty_texts (e, level, &left, &right);
+	orig = html_cursor_dup (e->cursor);
+	orig->position = position_before;
 	first = html_object_heads_list (obj);
 	last  = html_object_tails_list (obj);
-	set_cursor_at_end_of_object (e, obj, len);
+	set_cursor_at_end_of_object (e, obj, *len);
 
 	if ((left && left->data) || (right && (right->data))) {
 		HTMLObject *parent, *where;
@@ -645,12 +591,22 @@ insert_object_do (HTMLEngine *e, HTMLObject *obj, guint len, gboolean check, HTM
 			where  = NULL;
 			parent = HTML_OBJECT (right->data)->parent;
 		}
-		if (parent)
+		if (parent && html_object_is_clue (parent))
 			html_clue_append_after (HTML_CLUE (parent), obj, where);
 	}
 
+#ifdef OP_DEBUG
+	printf ("position before merge %d\n", e->cursor->position);
+#endif
 	remove_empty_and_merge (e, TRUE, last, right, orig);
 	remove_empty_and_merge (e, TRUE, left, first, orig);
+#ifdef OP_DEBUG
+	printf ("position after merge %d\n", e->cursor->position);
+#endif
+
+	html_cursor_destroy (e->cursor);
+	e->cursor = html_cursor_dup (orig);
+	html_cursor_jump_to_position_no_spell (e->cursor, e, position_after);
 
 	if (check)
 		html_engine_spell_check_range (e, orig, e->cursor);
@@ -662,23 +618,41 @@ struct _InsertUndo {
 	HTMLUndoData data;
 
 	guint len;
+	gboolean delete_paragraph_before;
+	gboolean delete_paragraph_after;
 };
 typedef struct _InsertUndo InsertUndo;
 
 static void
-insert_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir)
+insert_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection dir, guint position_after)
 {
 	InsertUndo *undo;
 
 	undo = (InsertUndo *) data;
 
 	html_engine_set_mark (e);
-	html_engine_move_cursor (e, HTML_ENGINE_CURSOR_LEFT, undo->len);
-	delete_object (e, NULL, NULL, html_undo_direction_reverse (dir));
+	html_cursor_jump_to_position (e->cursor, e, position_after);
+	delete_object (e, NULL, NULL, html_undo_direction_reverse (dir), TRUE);
+
+	if (undo->delete_paragraph_after || undo->delete_paragraph_before) {
+		html_cursor_jump_to_position (e->cursor, e, position_after);
+		if (undo->delete_paragraph_before) {
+			html_cursor_backward (e->cursor, e);
+		}
+		html_engine_set_mark (e);
+		if (undo->delete_paragraph_before) {
+			html_cursor_forward (e->cursor, e);
+		}
+		if (undo->delete_paragraph_after) {
+			html_cursor_forward (e->cursor, e);
+		}
+		delete_object (e, NULL, NULL, HTML_UNDO_UNDO, FALSE);
+	}
 }
 
 static void
-insert_setup_undo (HTMLEngine *e, guint len, HTMLUndoDirection dir)
+insert_setup_undo (HTMLEngine *e, guint len, guint position_before, HTMLUndoDirection dir,
+		   gboolean delete_paragraph_before, gboolean delete_paragraph_after)
 {
 	InsertUndo *undo;
 
@@ -686,31 +660,39 @@ insert_setup_undo (HTMLEngine *e, guint len, HTMLUndoDirection dir)
 
 	html_undo_data_init (HTML_UNDO_DATA (undo));
 	undo->len = len;
+	undo->delete_paragraph_before = delete_paragraph_before;
+	undo->delete_paragraph_after  = delete_paragraph_after;
 
 	/* printf ("insert undo len %d\n", len); */
 
 	html_undo_add_action (e->undo,
 			      html_undo_action_new ("Insert", insert_undo_action,
-						    HTML_UNDO_DATA (undo), html_cursor_get_position (e->cursor)),
+						    HTML_UNDO_DATA (undo),
+						    html_cursor_get_position (e->cursor),
+						    position_before),
 			      dir);
 }
 
 static void
-insert_object (HTMLEngine *e, HTMLObject *obj, guint len, HTMLUndoDirection dir, gboolean check)
+insert_object (HTMLEngine *e, HTMLObject *obj, guint len, guint position_after, gint level,
+	       HTMLUndoDirection dir, gboolean check)
 {
-	/* FIXME for tables */
-	if (HTML_IS_TABLE (obj))
-		append_object (e, obj, len, dir);
-	else if (len > 0) {
-		insert_object_do (e, obj, len, check, dir);
-		insert_setup_undo (e, len, dir);
-	}
+	gboolean delete_paragraph_before = FALSE;
+	gboolean delete_paragraph_after = FALSE;
+	guint position_before;
+
+	position_before = e->cursor->position;
+	insert_object_do (e, obj, &len, level, position_after, check, dir);
+	isolate_tables (e, dir, position_before, position_after, &delete_paragraph_before, &delete_paragraph_after);
+	html_cursor_jump_to_position_no_spell (e->cursor, e, position_after + (delete_paragraph_before ? 1 : 0));
+	insert_setup_undo (e, len, position_before + (delete_paragraph_before ? 1 : 0),
+			   dir, delete_paragraph_before, delete_paragraph_after);
 }
 
 void
-html_engine_insert_object (HTMLEngine *e, HTMLObject *o, guint len)
+html_engine_insert_object (HTMLEngine *e, HTMLObject *o, guint len, gint level)
 {
-	insert_object (e, o, len, HTML_UNDO_UNDO, TRUE);
+	insert_object (e, o, len, e->cursor->position + len, level, HTML_UNDO_UNDO, TRUE);
 }
 
 void
@@ -718,7 +700,7 @@ html_engine_paste_object (HTMLEngine *e, HTMLObject *o, guint len)
 {
 	html_undo_level_begin (e->undo, "Paste", "Paste");
 	html_engine_delete (e);
-	html_engine_insert_object (e, o, len);
+	html_engine_insert_object (e, o, len, html_object_get_insert_level (o));
 	html_undo_level_end (e->undo);
 }
 
@@ -737,19 +719,21 @@ html_engine_paste (HTMLEngine *e)
 static void
 check_magic_link (HTMLEngine *e, const gchar *text, guint len)
 {
-	if (HTML_OBJECT_TYPE (e->cursor->object) == HTML_TYPE_TEXT
+	if (HTML_IS_TEXT (e->cursor->object)
 	    && GTK_HTML_PROPERTY (e->widget, magic_links) && len == 1
 	    && (*text == ' ' || text [0] == '\n' || text [0] == '>' || text [0] == ')'))
 		html_text_magic_link (HTML_TEXT (e->cursor->object), e, html_object_get_length (e->cursor->object));
 }
 
 static void
-insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir)
+insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir, gboolean add_undo)
 {
 	GList *left=NULL, *right=NULL;
 	HTMLCursor *orig;
+	guint position_before;
 
 	html_engine_freeze (e);
+	position_before = e->cursor->position;
 	orig = html_cursor_dup (e->cursor);
 	split_and_add_empty_texts (e, 2, &left, &right);
 	remove_empty_and_merge (e, FALSE, left, right, orig);
@@ -765,7 +749,10 @@ insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir)
 		html_clue_append (HTML_CLUE (flow), e->cursor->object);
 	}
 
-	insert_setup_undo (e, 1, dir);
+	html_undo_level_begin (e->undo, "Insert paragraph", "Delete paragraph");
+	if (add_undo) {
+		insert_setup_undo (e, 1, position_before, dir, FALSE, FALSE);
+	}
 	g_list_free (left);
 	g_list_free (right);
 	html_engine_spell_check_range (e, orig, e->cursor);
@@ -774,19 +761,17 @@ insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir)
 	html_cursor_backward (e->cursor, e);
 	check_magic_link (e, "\n", 1);
 	html_cursor_forward (e->cursor, e);
+	
+	gtk_html_editor_event_command (e->widget, GTK_HTML_COMMAND_INSERT_PARAGRAPH, FALSE);
+	html_undo_level_end (e->undo);
 
 	html_engine_thaw (e);
-
-	gtk_html_editor_event_command (e->widget, GTK_HTML_COMMAND_INSERT_PARAGRAPH);
-
-	/* break links in new paragraph */
-	html_engine_insert_link (e, NULL, NULL);
 }
 
 void
 html_engine_insert_empty_paragraph (HTMLEngine *e)
 {
-	insert_empty_paragraph (e, HTML_UNDO_UNDO);
+	insert_empty_paragraph (e, HTML_UNDO_UNDO, TRUE);
 }
 
 void
@@ -800,6 +785,10 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 	if (!len)
 		return;
 
+	html_undo_level_begin (e->undo, "Insert text", "Delete text");
+	/* FIXME add insert text event */
+	gtk_html_editor_event_command (e->widget, GTK_HTML_COMMAND_INSERT_PARAGRAPH, TRUE);
+
 	do {
 		nl   = g_utf8_strchr (text, '\n');
 		alen = nl ? g_utf8_pointer_to_offset (text, nl) : len;
@@ -817,11 +806,14 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 			html_text_convert_nbsp (HTML_TEXT (o), TRUE);
 
 			if (alen == 1 && html_is_in_word (html_text_get_char (HTML_TEXT (o), 0))
-			    && !html_is_in_word (html_cursor_get_current_char (e->cursor)))
-					e->need_spell_check = TRUE;
-			else
+			    && !html_is_in_word (html_cursor_get_current_char (e->cursor))) {
+				/* printf ("need_spell_check\n"); */
+				e->need_spell_check = TRUE;
+			} else {
 				check = TRUE;
-			insert_object (e, o, html_object_get_length (o), HTML_UNDO_UNDO, check);
+			}
+			insert_object (e, o, html_object_get_length (o), e->cursor->position + html_object_get_length (o),
+				       1, HTML_UNDO_UNDO, check);
 		}
 		if (nl) {
 			html_engine_insert_empty_paragraph (e);
@@ -829,6 +821,7 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 			text = nl + 1;
 		}
 	} while (nl);
+	html_undo_level_end (e->undo);
 }
 
 void
@@ -852,13 +845,13 @@ html_engine_delete_container (HTMLEngine *e)
 	g_assert (e->cursor->object);
 	g_assert (html_object_is_container (e->cursor->object));
 
-	html_engine_freeze (e);
 	html_engine_set_mark (e);
+	html_engine_update_selection_if_necessary (e);
+	html_engine_freeze (e);
 	if (e->cursor->offset)
 		html_cursor_beginning_of_line (e->cursor, e);
 	else
 		html_cursor_end_of_line (e->cursor, e);
-	html_engine_update_selection_if_necessary (e);
 	html_engine_delete (e);
 	html_engine_thaw (e);
 }
@@ -871,6 +864,8 @@ html_engine_delete_n (HTMLEngine *e, guint len, gboolean forward)
 	else {
 		html_engine_block_selection (e);
 		html_engine_set_mark (e);
+		html_engine_update_selection_if_necessary (e);
+		html_engine_freeze (e);
 		while (len != 0) {
 			if (forward)
 				html_cursor_forward (e->cursor, e);
@@ -880,6 +875,7 @@ html_engine_delete_n (HTMLEngine *e, guint len, gboolean forward)
 		}
 		html_engine_delete (e);
 		html_engine_unblock_selection (e);
+		html_engine_thaw (e);	
 	}
 }
 
@@ -889,6 +885,7 @@ html_engine_cut_line (HTMLEngine *e)
 	g_return_if_fail (e != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (e));
 
+	html_undo_level_begin (e->undo, "Cut Line", "Undo Cut Line");
 	html_engine_set_mark (e);
 	html_engine_end_of_line (e);
 
@@ -896,6 +893,7 @@ html_engine_cut_line (HTMLEngine *e)
 		html_cursor_forward (e->cursor, e);
 
 	html_engine_cut (e);
+	html_undo_level_end (e->undo);
 }
 
 typedef struct {
@@ -922,7 +920,7 @@ change_link (HTMLObject *o, HTMLEngine *e, gpointer data)
 			html_clue_remove (HTML_CLUE (o->parent), o);
 			html_object_destroy (o);
 			if (changed->prev)
-				html_object_merge (changed->prev, changed, e, NULL, NULL);
+				html_object_merge (changed->prev, changed, e, NULL, NULL, NULL);
 		} else {
 			html_object_destroy (e->clipboard);
 			e->clipboard     = changed;
@@ -938,6 +936,8 @@ html_engine_set_insertion_link (HTMLEngine *e, const gchar *url, const gchar *ta
 	html_engine_set_target (e, target);
 	if (!url && e->insertion_color == html_colorset_get_color (e->settings->color_set, HTMLLinkColor))
 		html_engine_set_color (e, html_colorset_get_color (e->settings->color_set, HTMLTextColor));
+	else if (url)
+		html_engine_set_color (e, html_colorset_get_color (e->settings->color_set, HTMLLinkColor));
 }
 
 void
@@ -960,65 +960,147 @@ html_engine_insert_link (HTMLEngine *e, const gchar *url, const gchar *target)
 }
 
 static void
-append_object (HTMLEngine *e, HTMLObject *o, guint len, HTMLUndoDirection dir)
+prepare_empty_flow (HTMLEngine *e, HTMLUndoDirection dir)
 {
-	GList *left = NULL, *right = NULL;
-	HTMLObject *where;
-	gint back = 0;
-
-	html_engine_freeze (e);
-	if (html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))) {
-		HTMLObject *c, *cn;
-		HTMLClue *clue = HTML_CLUE (e->cursor->object->parent);
-		for (c = clue->head; c; c = cn) {
-			cn = c->next;
-			html_object_destroy (c);
-		}
-		clue->head = clue->tail = o;
-		e->cursor->object = o;
-		e->cursor->offset = 0;
-		o->parent = HTML_OBJECT (clue);
-	} else {
-		HTMLObject *flow;
-
-		flow  = html_clueflow_new (HTML_CLUEFLOW_STYLE_NORMAL, 0, HTML_LIST_TYPE_UNORDERED, 1);
-		html_clue_append (HTML_CLUE (flow), o);
-
-		html_object_split (e->cursor->object, e, NULL, e->cursor->offset, 2, &left, &right);
-		len += 2;
-		back = 1;
-
-		where = HTML_OBJECT (left->data);
-		html_clue_append_after (HTML_CLUE (where->parent), flow, where);
-
-		if (html_clueflow_is_empty (HTML_CLUEFLOW (where))) {
-			html_cursor_forward (e->cursor, e);
-			html_clue_remove (HTML_CLUE (where->parent), where);
-			html_object_destroy (where);
-			len --;
-			e->cursor->position --;
-		}
-		if (html_clueflow_is_empty (HTML_CLUEFLOW (HTML_OBJECT (flow)->next))) {
-			HTMLObject *empty;
-
-			empty = HTML_OBJECT (flow)->next;
-			html_clue_remove (HTML_CLUE (empty->parent), empty);
-			html_object_destroy (empty);
-			len --;
-			back = 0;
+	if (!html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))) {
+		insert_empty_paragraph (e, dir, TRUE);
+		if (e->cursor->object->parent->prev
+		    && html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent->prev))) {
+			html_cursor_backward (e->cursor, e);
+		} else if (!html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))) {
+			insert_empty_paragraph (e, dir, TRUE);
+			html_cursor_backward (e->cursor, e);
 		}
 	}
+}
+
+static void
+append_object (HTMLEngine *e, HTMLObject *o, guint len, HTMLUndoDirection dir)
+{
+	HTMLObject *c, *cn;
+	HTMLClue *clue;
+	guint position_before;
+
+	html_engine_freeze (e);
+	prepare_empty_flow (e, dir);
+	position_before = e->cursor->position;
+
+	g_return_if_fail (html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent)));
+
+	clue = HTML_CLUE (e->cursor->object->parent);
+	for (c = clue->head; c; c = cn) {
+		cn = c->next;
+		html_object_destroy (c);
+	}
+	clue->head = clue->tail = o;
+	e->cursor->object = o;
+	e->cursor->offset = 0;
+	o->parent = HTML_OBJECT (clue);
 
 	html_cursor_forward_n (e->cursor, e, len);
 	html_object_change_set (o, HTML_CHANGE_ALL_CALC);
 	html_engine_thaw (e);
 
-	insert_setup_undo (e, len, dir);
-	html_cursor_backward_n (e->cursor, e, back);
+	insert_setup_undo (e, len, position_before, dir, FALSE, FALSE);
+
+	return;
 }
 
 void
 html_engine_append_object (HTMLEngine *e, HTMLObject *o, guint len)
 {
+	html_undo_level_begin (e->undo, "Append object", "Remove appended object");
 	append_object (e, o, len, HTML_UNDO_UNDO);
+	html_undo_level_end (e->undo);
+}
+
+static void
+append_flow (HTMLEngine *e, HTMLObject *o, guint len, HTMLUndoDirection dir)
+{
+	HTMLObject *where;
+	guint position, position_before;
+
+	html_engine_freeze (e);
+
+	position_before = e->cursor->position;
+	prepare_empty_flow (e, dir);
+
+	g_return_if_fail (html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent)));
+
+	where = e->cursor->object->parent;
+
+	e->cursor->object = html_object_get_head_leaf (o);
+	e->cursor->offset = 0;
+	position = e->cursor->position;
+	while (html_cursor_backward (e->cursor, e))
+		;
+	e->cursor->position = position;
+	html_clue_append_after (HTML_CLUE (where->parent), o, where);
+	html_object_remove_child (where->parent, where);
+	html_object_destroy (where);
+
+	html_cursor_forward_n (e->cursor, e, len);
+	html_object_change_set (o, HTML_CHANGE_ALL_CALC);
+	html_engine_thaw (e);
+
+	insert_setup_undo (e, len, position_before, dir, FALSE, FALSE);
+
+	return;
+}
+
+void
+html_engine_append_flow (HTMLEngine *e, HTMLObject *o, guint len)
+{
+	html_undo_level_begin (e->undo, "Append flow", "Remove appended flow");
+	append_flow (e, o, len, HTML_UNDO_UNDO);
+	html_undo_level_end (e->undo);
+}
+
+void
+html_engine_cut_and_paste_begin (HTMLEngine *e, const gchar *undo_op_name, const gchar *redo_op_name)
+{
+	guint position;
+	gint level;
+
+	html_engine_hide_cursor (e);
+	html_engine_selection_push (e);
+	html_engine_clipboard_push (e);
+	html_undo_level_begin (e->undo, undo_op_name, redo_op_name);
+	position = e->mark ? MAX (e->cursor->position, e->mark->position) : e->cursor->position;
+	level = html_engine_cut (e);
+
+	e->cut_and_paste_stack = g_list_prepend (e->cut_and_paste_stack, GINT_TO_POINTER (level));
+	e->cut_and_paste_stack = g_list_prepend (e->cut_and_paste_stack, GUINT_TO_POINTER (position));
+}
+
+void
+html_engine_cut_and_paste_end (HTMLEngine *e)
+{
+	guint position;
+	gint level;
+
+	position = GPOINTER_TO_UINT (e->cut_and_paste_stack->data);
+	e->cut_and_paste_stack = g_list_remove (e->cut_and_paste_stack, e->cut_and_paste_stack->data);
+	level    = GPOINTER_TO_INT (e->cut_and_paste_stack->data);
+	e->cut_and_paste_stack = g_list_remove (e->cut_and_paste_stack, e->cut_and_paste_stack->data);
+
+	if (e->clipboard) {
+		insert_object (e, e->clipboard, e->clipboard_len, position, level, HTML_UNDO_UNDO, TRUE);
+		e->clipboard = NULL;
+	}
+	html_undo_level_end (e->undo);
+	html_engine_clipboard_pop (e);
+	html_engine_selection_pop (e);
+	html_engine_show_cursor (e);
+}
+
+void
+html_engine_cut_and_paste (HTMLEngine *e, const gchar *undo_op_name, const gchar *redo_op_name,
+			   HTMLObjectForallFunc iterator, gpointer data)
+{
+	html_engine_edit_selection_updater_update_now (e->selection_updater);
+	html_engine_cut_and_paste_begin (e, undo_op_name, redo_op_name);
+	if (e->clipboard)
+		html_object_forall (e->clipboard, e, iterator, data);
+	html_engine_cut_and_paste_end (e);
 }
